c$Id:$
      subroutine usolve(flags,b)

c      * * F E A P * * A Finite Element Analysis Program

c....  Copyright (c) 1984-2004: Regents of the University of California
c                               All rights reserved

c-----[--.----+----.----+----.-----------------------------------------]
c     Purpose:  Solver interface for WSMP

c     Inputs:
c        flags(1) - Allocation and/or initialization phase
c        flags(2) - Perform factorization for direct solutions
c        flags(3) - Coefficient array unsymmetric
c        flags(4) - Solve equations
c        flags(5) - Purge storage of pointers
c        b(*)     - RHS vector

c     Outputs:

c        flags(5) - True if error occurs (for factor/solve only)
c-----[--.----+----.----+----.-----------------------------------------]

      implicit  none

      include  'cdata.h'
      include  'compac.h'
      include  'debugs.h'
      include  'endata.h'
      include  'eqsym.h'
      include  'iofile.h'
      include  'part0.h'
      include  'setups.h'
      include  'pointer.h'
      include  'comblk.h'

      include  'wsmp_com.h'

      character tname*5
      logical   flags(*), setvar,palloc
      integer   i,ldb,mrp, naux,nnu,nrhs,ibsv, iparm(64)
      real*8    b(*), aux, dot, dparm(64)
      real*8    diag

      save

      data      aux / 0.0d0 /, diag / 0.0d0 /

c     Set number right hand sides

      ldb  = neq
      mrp  = 0
      naux = 0
      nrhs = 1

c     Presolve setups

      if(flags(1)) then

c       Check if problem is symmetric or unsymmetric

        if(flags(3)) then
          neqs = 1
          uall = .true.       ! Assemble total array
        else
          neqs = neq
          uall = .false.      ! Assemble upper triangular part
        endif

        ubycol = .false.      ! Store by rows
        udiag  = .true.       ! Diagonal in sparse array
        ulfl   = .false.      ! Prevents assembly of al in cassem

c       Establish current matrix indices

        call iters(nnu,-2)

c       Allocate matrix storage

        write(tname,'(4hTANG,i1)') npart
        setvar = palloc(npart,tname,nnu+1,2)
        call pzero(hr(np(npart)),nnu+1)

c       Adjust the pointer array

        setvar = palloc(227,'USOL3',neq+1,1)
        mr(np(227)) = 1
        do i = 1,neq
          mr(np(227)+i) = mr(np(225)+i-1) + 1
        end do ! i

c       Fill 'iparm' and 'dparm' arrays with default values.

        iparm(1) = 0
        iparm(2) = 0
        iparm(3) = 0
        if(uall) then
          if(wsmp_t.eq.1) then
            if(debug) then
              write(*,*) ' CLEAR SYMMETRIC DATA ',wsmp_t
            endif
            call wsmpclear()
          endif
          wsmp_t = 2
          call wgsmp (neq, mr(np(227)), mr(np(226)), hr(np(npart)),
     &                b, ldb, nrhs, aux, iparm, dparm)
        else
          if(wsmp_t.eq.2) then
            if(debug) then
              write(*,*) ' CLEAR UNSYMMETRIC DATA ',wsmp_t,naux
            endif
            call wsmpclear()
          endif
          wsmp_t = 1
          naux   = 0
          call wssmp (neq, mr(np(227)), mr(np(226)), hr(np(npart)),
     &                diag, mr(np(47)),mr(np(48)), b, ldb, nrhs,
     &                aux, naux, mrp, iparm, dparm)
        endif

        if (iparm(64) .ne. 0) then
          write(*,*) 'The following ERROR was detected: ',iparm(64)
          call plstop()
        end if

c     Solution steps for assembled equations

      else

c       Factor equations

        if(flags(2)) then

c         Ordering and Factorization

          if(uall) then
            iparm(2)  = 1
            iparm(3)  = 2
            call wgsmp (neq, mr(np(227)), mr(np(226)), hr(np(npart)),
     &                  b, ldb, nrhs, aux, iparm, dparm)
          else
            iparm(2)  = 1
            iparm(3)  = 3
            iparm(31) = 1
            call wssmp (neq, mr(np(227)), mr(np(226)), hr(np(npart)),
     &                  diag, mr(np(47)),mr(np(48)), b, ldb, nrhs,
     &                  aux, naux, mrp, iparm, dparm)
          endif

          if (iparm(64) .ne. 0) then
            write(*,*) 'The following ERROR was detected: ',iparm(64)
            call plstop()
          end if

          if(debug) then
            write(*,*) 'Number of nonzeros in factor L   = ',iparm(24)
            write(*,*) 'Number of FLOPS in factorization = ',dparm(23)
            write(*,*) 'Double words needed to factor    = ',iparm(23)

            write(iow,*)'Number of nonzeros in factor L   = ',iparm(24)
            write(iow,*)'Number of FLOPS in factorization = ',dparm(23)
            write(iow,*)'Double words needed to factor    = ',iparm(23)
          endif

        endif

c       Perform solve

        if(flags(4)) then

          setvar = palloc(228,'USOL4',neq  ,2)
          ibsv   = np(228) - 1
          do i = 1,neq
            hr(ibsv+i) = b(i)
          end do ! i

c         Solve system using LDL' factorization.

          if(uall) then
            iparm(2)  = 3
            iparm(3)  = 3
            call wgsmp (neq, mr(np(227)), mr(np(226)), hr(np(npart)),
     &                  b, ldb, nrhs, aux, iparm, dparm)
          else
            iparm(2)  = 4
            iparm(3)  = 4
            iparm(30) = 0
            iparm(31) = 1
            call wssmp (neq, mr(np(227)), mr(np(226)), hr(np(npart)),
     &                  diag, mr(np(47)),mr(np(48)), b, ldb, nrhs,
     &                  aux, naux, mrp, iparm, dparm)
          endif

          if (iparm(64) .ne. 0) then
            write(*,*) 'The following ERROR was detected: ',iparm(64)
            call plstop()
          end if

c         Compute 'energy' for convergence test

          aengy  = dot(b,hr(np(228)),neq)
          setvar = palloc(228,'USOL4',    0,2)

        endif

c       Purge storage in 'factor'

        if(flags(5)) then
          setvar = palloc(227,'USOL3',    0,1)
          setvar = palloc(226,'USOL2',    0,1)  ! allocated in uiters
          setvar = palloc(225,'USOL1',    0,1)  ! allocated in uiters
        endif

      endif

      end
